package org.dcache.namespace;

import java.util.EnumSet;
import java.util.Set;
import java.util.Collections;
import javax.security.auth.Subject;

import org.dcache.auth.Subjects;
import org.dcache.vehicles.FileAttributes;

import org.dcache.acl.ACL;
import org.dcache.acl.Permission;
import org.dcache.acl.Owner;
import org.dcache.acl.mapper.AclMapper;
import org.dcache.acl.matcher.AclNFSv4Matcher;
import org.dcache.acl.enums.AccessType;
import org.dcache.auth.Origin;
import static org.dcache.acl.enums.AccessType.*;
import static org.dcache.namespace.FileAttribute.*;
import static org.dcache.acl.enums.AccessMask.*;

/**
 * A PermissionHandler using the ACL module as a PDP.
 */
public class ACLPermissionHandler implements PermissionHandler
{
    private Set<FileAttribute> _requiredAttributes =
        Collections.unmodifiableSet(EnumSet.of(ACL, OWNER, OWNER_GROUP));

    @Override
    public Set<FileAttribute> getRequiredAttributes()
    {
        return _requiredAttributes;
    }

    private Permission getPermission(Subject subject, FileAttributes attr)
    {
        ACL acl = attr.getAcl();
        Owner owner = new Owner(attr.getOwner(), attr.getGroup());
        Origin origin = Subjects.getOrigin(subject);
        return AclMapper.getPermission(subject, origin, owner, acl);
    }

    @Override
    public AccessType canReadFile(Subject subject, FileAttributes attr)
    {
        Permission permission = getPermission(subject, attr);
        return valueOf(AclNFSv4Matcher.isAllowed(permission, READ_DATA));
    }

    @Override
    public AccessType canWriteFile(Subject subject, FileAttributes attr)
    {
        Permission permission = getPermission(subject, attr);
        return valueOf(AclNFSv4Matcher.isAllowed(permission, WRITE_DATA));
    }

    @Override
    public AccessType canCreateSubDir(Subject subject, FileAttributes attr)
    {
        Permission permission = getPermission(subject, attr);
        return valueOf(AclNFSv4Matcher.isAllowed(permission, ADD_SUBDIRECTORY));
    }

    @Override
    public AccessType canCreateFile(Subject subject, FileAttributes attr)
    {
        Permission permission = getPermission(subject, attr);
        return valueOf(AclNFSv4Matcher.isAllowed(permission, ADD_FILE));
    }

    @Override
    public AccessType canDeleteFile(Subject subject,
                                    FileAttributes parentAttr,
                                    FileAttributes childAttr)
    {
        Permission permissionParent = getPermission(subject, parentAttr);
        AccessType ofParent =  valueOf(AclNFSv4Matcher.isAllowed(permissionParent,
                                                            DELETE_CHILD));
        if ( ofParent == ACCESS_ALLOWED )
            return ofParent;

        Permission permissionChild = getPermission(subject, childAttr);
        AccessType ofChild =  valueOf(AclNFSv4Matcher.isAllowed(permissionChild,
                                                            DELETE));

        if (ofChild == ACCESS_ALLOWED)
            return ofChild;

        if (ofParent == ACCESS_DENIED
                || ofChild == ACCESS_DENIED) {
            return ACCESS_DENIED;
        }

        return ACCESS_UNDEFINED;
    }

    @Override
    public AccessType canDeleteDir(Subject subject,
                                   FileAttributes parentAttr,
                                   FileAttributes childAttr)
    {
        return canDeleteFile(subject, parentAttr, childAttr);
    }

    @Override
    public AccessType canListDir(Subject subject, FileAttributes attr)
    {
        Permission permission = getPermission(subject, attr);
        return valueOf(AclNFSv4Matcher.isAllowed(permission,
                                                            LIST_DIRECTORY));
    }

    @Override
    public AccessType canLookup(Subject subject, FileAttributes attr)
    {
        Permission permission = getPermission(subject, attr);
        return valueOf(AclNFSv4Matcher.isAllowed(permission,
                                                            EXECUTE));
    }

    @Override
    public AccessType canRename(Subject subject,
                                FileAttributes parentAttr,
                                FileAttributes newParentAttr,
                                boolean isDirectory)
    {
        Permission permission1 = getPermission(subject, parentAttr);
        Permission permission2 = getPermission(subject, newParentAttr);

        Boolean ofSrcParent = AclNFSv4Matcher.isAllowed(permission1, DELETE_CHILD);
        Boolean ofDestParent = AclNFSv4Matcher.isAllowed(permission2, ADD_FILE);

        if (ofDestParent == ofSrcParent)
            return valueOf(ofSrcParent);

        if (valueOf(ofSrcParent) == ACCESS_DENIED ||
                valueOf(ofDestParent) == ACCESS_DENIED)
            return ACCESS_DENIED;

        return ACCESS_UNDEFINED;
    }

    @Override
    public AccessType canSetAttributes(Subject subject,
                                       FileAttributes parentAttr,
                                       FileAttributes attr,
                                       Set<FileAttribute> attributes)
    {
        Permission permission = getPermission(subject, attr);
        return canSetAttributes(permission, attributes);
    }

    @Override
    public AccessType canGetAttributes(Subject subject,
                                       FileAttributes parentAttr,
                                       FileAttributes attr,
                                       Set<FileAttribute> attributes)
    {
        Permission permission = getPermission(subject, attr);
        return canGetAttributes(permission, attributes);
    }

    /**
     * Determines if the action is allowed on the get of file
     * attributes. Returns ACCESS_DENIED if the action is denied for
     * one or more of the attributes. Returns ACCESS_ALLOWED if the
     * action is allowed for all attributes. Returns ACCESS_UNDEFINED
     * otherwise.
     */
    private AccessType canGetAttributes(Permission permission, Set<FileAttribute> attributes)
    {
        if (attributes.contains(ACL) ) {
            return valueOf(AclNFSv4Matcher.isAllowed(permission, READ_ACL));
        }

        return valueOf(AclNFSv4Matcher.isAllowed(permission, READ_ATTRIBUTES));
    }

    /**
     * Determines if the action is allowed on the set of file attributes.
     * Returns ACCESS_DENIED if the action is denied for one or more of the
     * attributes. Returns ACCESS_ALLOWED if the action is allowed for all
     * attributes. Returns ACCESS_UNDEFINED otherwise.
     */
    private AccessType canSetAttributes(Permission permission, Set<FileAttribute> attributes) {
        if (attributes.contains(ACL)) {
            return valueOf(AclNFSv4Matcher.isAllowed(permission, WRITE_ACL));
        }

        return valueOf(AclNFSv4Matcher.isAllowed(permission, WRITE_ATTRIBUTES));
    }
}