/*
COPYRIGHT STATUS:
Dec 1st 2001, Fermi National Accelerator Laboratory (FNAL) documents and
software are sponsored by the U.S. Department of Energy under Contract No.
DE-AC02-76CH03000. Therefore, the U.S. Government retains a  world-wide
non-exclusive, royalty-free license to publish or reproduce these documents
and software for U.S. Government purposes.  All documents and software
available from this server are protected under the U.S. and Foreign
Copyright Laws, and FNAL reserves all rights.

Distribution of the software available from this server is free of
charge subject to the user following the terms of the Fermitools
Software Legal Information.

Redistribution and/or modification of the software shall be accompanied
by the Fermitools Software Legal Information  (including the copyright
notice).

The user is asked to feed back problems, benefits, and/or suggestions
about the software to the Fermilab Software Providers.

Neither the name of Fermilab, the  URA, nor the names of the contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

DISCLAIMER OF LIABILITY (BSD):

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED  WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED  WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL FERMILAB,
OR THE URA, OR THE U.S. DEPARTMENT of ENERGY, OR CONTRIBUTORS BE LIABLE
FOR  ANY  DIRECT, INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT  OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE  POSSIBILITY OF SUCH DAMAGE.

Liabilities of the Government:

This software is provided by URA, independent from its Prime Contract
with the U.S. Department of Energy. URA is acting independently from
the Government and in its own private capacity and is not acting on
behalf of the U.S. Government, nor as its contractor nor its agent.
Correspondingly, it is understood and agreed that the U.S. Government
has no connection to this software and in no manner whatsoever shall
be liable for nor assume any responsibility or obligation for any claim,
cost, or damages arising out of or resulting from the use of the software
available from this server.

Export Control:

All documents and software available from this server are subject to U.S.
export control laws.  Anyone downloading information from this server is
obligated to secure any necessary Government licenses before exporting
documents or software obtained from this server.
 */
package org.dcache.webadmin.model.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.jdo.FetchPlan;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import org.dcache.alarms.Severity;
import org.dcache.alarms.dao.AlarmEntry;

import com.google.common.base.Strings;
import com.google.common.base.Preconditions;

/**
 * Convenience methods for generating JDO queries from {@link AlarmEntry}
 * fields.
 *
 * @author arossi
 */
public class AlarmJDOUtils {

    public static class AlarmDAOFilter {
        private String filter;
        private String parameters;
        private Object[] values;

        @Override
        public String toString() {
            return filter
                            + ", "
                            + parameters
                            + (values == null ? null : ", "
                                            + Arrays.asList(values));
        }

        public String getFilter() {
            return filter;
        }

        public String getParameters() {
            return parameters;
        }

        public Object[] getValues() {
            return values;
        }
    }

    /**
     * Construct an actual JDO query from the filter.
     */
    public static Query createQuery(PersistenceManager pm, AlarmDAOFilter filter) {
        String expression;
        String parameters;

        if (filter != null) {
            expression = filter.filter;
            parameters = filter.parameters;
        } else {
            expression = null;
            parameters = null;
        }

        if (parameters == null) {
            if (expression == null) {
                return pm.newQuery(AlarmEntry.class);
            }
            return pm.newQuery(AlarmEntry.class, expression);
        }
        Query query = pm.newQuery(AlarmEntry.class);
        query.setFilter(expression);
        query.declareParameters(parameters);
        query.addExtension("datanucleus.query.resultCacheType", "none");
        query.getFetchPlan().setFetchSize(FetchPlan.FETCH_SIZE_OPTIMAL);
        return query;
    }

    public static long delete(PersistenceManager pm, AlarmDAOFilter filter) {
        Query query = AlarmJDOUtils.createQuery(pm, filter);
        return filter.values == null ? query.deletePersistentAll()
                        : query.deletePersistentAll(filter.values);
    }

    public static Collection<AlarmEntry> execute(PersistenceManager pm,
                    AlarmDAOFilter filter) {
        Query query = AlarmJDOUtils.createQuery(pm, filter);
        return (Collection<AlarmEntry>) (filter.values == null ? query.execute()
                        : query.executeWithArray(filter.values));
    }

    /**
     * Construct filter based on values for the before and closed (AND'd).
     */
    public static AlarmDAOFilter getDeleteBeforeFilter(Long before) {
        Preconditions.checkNotNull(before);
        AlarmDAOFilter filter = new AlarmDAOFilter();
        filter.filter = "timestamp<=b && closed==t";
        filter.parameters = "java.lang.Long b, java.lang.Boolean t";
        filter.values = new Object[]{before, true};
        return filter;
    }

    /**
     * Construct filter based on values for the parameter fields (AND'd).
     */
    public static AlarmDAOFilter getFilter(Date after, Date before,
                    Severity severity, String type) {
        StringBuilder f = new StringBuilder();
        StringBuilder p = new StringBuilder();
        List<Object> values = new ArrayList<>();

        if (after != null) {
            f.append("timestamp>=a");
            p.append("java.lang.Long a");
            values.add(after.getTime());
        }

        if (before != null) {
            if (f.length() > 0) {
                f.append(" && ");
                p.append(", ");
            }
            f.append("timestamp<=b");
            p.append("java.lang.Long b");
            values.add(before.getTime());
        }

        if (severity != null) {
            if (f.length() > 0) {
                f.append(" && ");
                p.append(", ");
            }
            f.append("severity>=s");
            p.append("java.lang.Integer s");
            values.add(severity.ordinal());
        }

        if (type != null) {
            if (f.length() > 0) {
                f.append(" && ");
                p.append(", ");
            }
            f.append("type==t");
            p.append("java.lang.String t");
            values.add(type);
        }

        AlarmDAOFilter filter = new AlarmDAOFilter();
        filter.filter = trimToNull(f);
        filter.parameters = trimToNull(p);
        filter.values = emptyListToNull(values);
        return filter;
    }

    /**
     * Construct filter from multiple alarm entry keys (OR'd).
     */
    public static AlarmDAOFilter getIdFilter(Collection<AlarmEntry> selected) {
        StringBuilder f = new StringBuilder();
        StringBuilder p = new StringBuilder();
        List<Object> values = new ArrayList<>();
        Iterator<AlarmEntry> i = selected.iterator();
        int k = 0;

        if (i.hasNext()) {
            f.append("key==k").append(k);
            p.append("java.lang.String k").append(k++);
            values.add(i.next().getKey());
        }

        /*
         * create successive 'or' clauses if any
         */
        while (i.hasNext()) {
            f.append(" || key==k").append(k);
            p.append(", java.lang.String k").append(k++);
            values.add(i.next().getKey());
        }

        AlarmDAOFilter filter = new AlarmDAOFilter();
        filter.filter = Strings.emptyToNull(f.toString());
        filter.parameters = Strings.emptyToNull(p.toString());
        filter.values = emptyListToNull(values);
        return filter;
    }

    private static Object[] emptyListToNull(List<Object> values) {
        return values.isEmpty() ? null : values.toArray();
    }

    private static String trimToNull(StringBuilder buffer) {
        String trimmed = buffer.toString().trim();
        if (trimmed.length() == 0) {
            return null;
        }
        return trimmed;
    }

    private AlarmJDOUtils() {
    }
}
